import { join, values } from '../fn.js'
import { alpha, alphanumeric, any, char, digit, many, map, maybe, noCaseString, seq, skip, string, until } from '../parser.js'
import { Alias, Identifier, Literal, ObjectPath, Property } from './types.js'

export const Symbols = Object.freeze({
  Bracket: Object.freeze({
    Angle: Object.freeze({
      Left: char('<'),
      Right: char('>')
    }),
    Curly: Object.freeze({
      Left: char('{'),
      Right: char('}'),
    }),
    Round: Object.freeze({
      Left: char('('),
      Right: char(')'),
    }),
    Square: Object.freeze({
      Left: char('['),
      Right: char(']'),
    }),
  }),
  Colon: char(':'),
  Comma: char(','),
  Equal: char('='),
  Hyphen: char('-'),
  Newline: char('\n'),
  Tab: char('\t'),
  Period: char('.'),
  Plus: char('+'),
  Space: char(' '),
  Quote: char('"')
})

export const Keyword = Object.freeze({
  Use: noCaseString('use'),
  Match: noCaseString('match'),
  Where: noCaseString('where'),
  Return: noCaseString('return')
})

export const word = value => seq(ws, value, ws)
export const collect = parser => map(join(''), parser)
export const quoted = collect(seq(skip(Symbols.Quote), until(Symbols.Quote), skip(Symbols.Quote)))
export const number = seq(digit, many(digit))
export const signed = seq(maybe(any(Symbols.Plus, Symbols.Hyphen)), number)
export const ws = skip(many(any(Symbols.Newline, Symbols.Space, Symbols.Tab)))
export const identifier = map(([x]) => new Identifier(x), collect(seq(alpha, many(alphanumeric))))

const float = map(([n]) => parseFloat(n, 10), collect(seq(signed, Symbols.Period, number)))
const integer = map(([n]) => parseInt(n, 10), collect(signed))
const str = quoted

const toBoolean = v => v === 'false' ? false : true
const boolean = map(([v]) => toBoolean(v), any(string('true'), string('false')))

const property = map(([x]) => new Property(x), seq(skip(Symbols.Period), identifier))
const accessor = map(x => new ObjectPath(...x), seq(identifier, property, many(property)))

export const literal = map(([x]) => new Literal(x), any(float, integer, str, boolean))
export const baseValue = any(literal, accessor, identifier)

const as = noCaseString('as')
export const alias = map(([value, name]) => new Alias(value, name), seq(baseValue, word(skip(as)), identifier))
export const value = any(alias, baseValue)
export const keyword = any(...values(Keyword))

