import { defineQuery, hasComponent } from 'bitecs'
import { query as q } from './query-parser/index.js'
import { parseAll } from './parser.js'
import { curry, of, is, map, always, assocPath, prop, identity, mergeRightDeep, pipe, filter, some, path, tap } from './fn.js'
import { Relationship } from './query-parser/types.js'

const nodeComponent = component => node => {
  const label = node.label?.valueOf()
  const name = node.name?.valueOf()
  const type = label && component[label] || identity

  return { name, label, type, properties: node.properties }
}

const notEntity = ({ label }) => label != null && label.valueOf() !== 'Entity'

const definePropertyQuery = components => {
  const query = pipe(
    filter(notEntity),
    map(prop('type')),
    defineQuery
  )(components)

  if (some(c => c.properties.length > 0, components)) {
    return world => {
      const comps = filter(c => c.properties.length > 0, components)
      return query(world).filter(eid =>
        comps.every(cmp =>
          cmp.properties.every(({ key, value }) =>
            cmp.type[key][eid] === value.valueOf()
          )
        )
      )
    }
  } else {
    return query
  }
}

const prepareQuery = (match, component) => {
  const relationship = is(Relationship, match)

  if (relationship) {
    const { from: f, to: t, edge: e } = match
    const [from, to, edge] = map(match => prepareQuery(match, component), [f, t, e])

    return { from, to, edge, relationship, schema: query }
  } else {
    const components = match.components.map(nodeComponent(component))
    // const types = components.filter(notEntity).map(prop('type'))
    const query = definePropertyQuery(components)
    return { components, relationship, query, schema: match }
  }
}

const hasComponents = (world, types, entity) => (
  types.every(type => hasComponent(world, type, entity))
)

const matches = (world, node, results, set = new Set()) => {
  const types = node.components.filter(notEntity).map(prop('type'))
  return entity => {
    if (set.has(entity)) {
      return true
    } else if (hasComponents(world, types, entity) && (results == null || results.includes(entity))) {
      set.add(entity)
      return true
    } else {
      return false
    }
  }
}

const queryRelationship = (query, world) => {
  const [from, edge, to] = ['from', 'edge', 'to'].map(name => {
    const node = query[name]
    return node.relationship ? node.from : node
  })

  //const [fc, tc] = [from, to].map(n => n.components.filter(notEntity).map(prop('type')))
  const matchesFrom = matches(world, from, from.query(world))
  const matchesTo = matches(world, to)

  const Edge = edge.components[0].type
  const edges = edge.query(world)
  const result = []
  for (let i = 0; i < edges.length; i++) {
    const eid = edges[i]
    if (matchesFrom(Edge.from[eid]) && matchesTo(Edge.to[eid])) {
      result.push(eid)
    }
  }

  return result
}

const assembleQuery = (query, entities) => map(entity => ([{ entity, query }]), entities)

const executeQuery = curry((query, world) => {
  if (!query.relationship) {
    return assembleQuery(query, query.query(world))
  }

  const edges = queryRelationship(query, world)

  if (query.to.relationship) {
    return edges.reduce((acc, eid) => {
      const next = assocPath(
        'from.query',
        always(of(query.edge.components[0].type.to[eid])),
        query.to
      )

      return executeQuery(next, world)
        .map(child => [eid, ...child])
        .map(([l, r]) => [{ entity: l, query }, r])
        .concat(acc)
    }, [])
  } else {
    return assembleQuery(query, edges)
  }
})

const isTag = (type) => (
  type && Object.getOwnPropertySymbols(type).find(
    (s) => s.description === 'tagStore'
  ) || false
)

const getValue = curry((type, entity) => {
  if (isTag(type)) {
    return {}
  } else if (typeof type === 'function') {
    return type(entity)
  } else {
    return Object.fromEntries(
      Object.entries(type).map(([k, v]) => [k, v[entity]])
    )
  }
})

const resolveNode = node => {
  const values = node.query.components.map(({ name, type }) => ({ [name]: getValue(type, node.entity) }))
  return {
    ...node,
    values: Object.assign({}, ...values)
  }
}

const resolveRelationship = edges => {
  const { entity, query } = edges
  const Edge = query.edge.components[0].type

  const to = query.to?.from ?? query.to
  return [
    resolveNode({ entity: Edge.from[entity], query: query.from }),
    resolveNode({ entity, query: query.edge }),
    resolveNode({ entity: Edge.to[entity], query: to }),
  ]
}

const resolveValues = query => {
  return query.map(edges => {
    const values = edges.map(edge => {
      if (edge.query.relationship) {
        return resolveRelationship(edge)
      } else {
        return resolveNode(edge, {})
      }
    })

    return values.flat().reduce(mergeRightDeep)
  })
}

const filterValues = curry((filters, values) => {
  return values.filter(v => filters.every(f => f.passes(v)))
})

const resolveReturns = curry((returnValues, values) => {
  return values.map(x => returnValues.from(x.values))
})

export const query = curry((input, { world, component }) => {
  return parseAll(q, input).map(({ use, matches, filters, returnValues, ...rest }) => {
    const useWorld = world[use?.valueOf() ?? 'default']
    const match = matches[0].valueOf()
    const preparedQuery = prepareQuery(match, component)
    const results = executeQuery(preparedQuery, useWorld)
    const values = resolveValues(results)
    const filtered = filterValues(filters, values)
    const returns = resolveReturns(returnValues, filtered)
    return returns
  })
})

